# -*- coding: utf-8 -*-
# pragma pylint: disable=wildcard-import,redefined-builtin
"""Contains all typet classes and functions."""

from __future__ import unicode_literals

from .meta import *  # noqa: F401
from .objects import *  # noqa: F401
from .path import *  # noqa: F401
from .types import *  # noqa: F401
from .validation import *  # noqa: F401



# -*- coding: utf-8 -*-
"""A module containing common metaclasses and utilites for working with them.

Metaclasses:
    Singleton: A metaclass to force a class to only ever be instantiated once.
    IdempotentSingleton: A metaclass that will force a class to only create one
        instance, but will call __init__ on the instance when new instantiation
        attempts occur.
    Uninstantiable: A metaclass that causes a class to be uninstantiable.

Decorators:
    metaclass: A class decorator that will create the class using multiple
        metaclasses.
    singleton: A class decorator that will make the class a singleton, even if
        the class already has a metaclass.
"""

from __future__ import unicode_literals

import collections

from typingplus import (  # noqa: F401 pylint: disable=unused-import
    Any,
    Callable
)
import six


__all__ = (
    'metaclass',
    'Singleton',
    'singleton',
    'IdempotentSingleton',
    'Uninstantiable',
)


def metaclass(*metaclasses):
    # type: (*type) -> Callable[[type], type]
    """Create the class using all metaclasses.

    Args:
        metaclasses: A tuple of metaclasses that will be used to generate and
            replace a specified class.

    Returns:
        A decorator that will recreate the class using the specified
        metaclasses.
    """
    def _inner(cls):
        # pragma pylint: disable=unused-variable
        metabases = tuple(collections.OrderedDict(  # noqa: F841
            (c, None) for c in (metaclasses + (type(cls),))
        ).keys())
        # pragma pylint: enable=unused-variable
        _Meta = metabases[0]
        for base in metabases[1:]:
            class _Meta(base, _Meta):  # pylint: disable=function-redefined
                pass
        return six.add_metaclass(_Meta)(cls)
    return _inner


class Singleton(type):
    """A metaclass to turn a class into a singleton.

    If the instance already exists, Singleton will attempt to call
    __singleton__ on the instance to allow the instance to update if necessary.
    """

    __instance__ = None  # type: type

    def __call__(cls, *args, **kwargs):
        # type: (*Any, **Any) -> type
        """Instantiate the class only once."""
        if not cls.__instance__:
            cls.__instance__ = super(Singleton, cls).__call__(*args, **kwargs)
        else:
            try:
                cls.__instance__.__singleton__(*args, **kwargs)  # type: ignore
            except (AttributeError, TypeError):
                pass
        return cls.__instance__


class IdempotentSingleton(Singleton):
    """A metaclass to turn a class into a singleton.

    If the instance already exists, IdempotentSingleton will call __init__ on
    the existing instance with the arguments given.
    """

    def __call__(cls, *args, **kwargs):
        # type: (*Any, **Any) -> type
        """Create one instance of the class and reinstantiate as necessary."""
        if not cls.__instance__:
            cls.__instance__ = super(IdempotentSingleton, cls).__call__(
                *args, **kwargs)
        else:
            try:
                cls.__instance__.__init__(*args, **kwargs)  # type: ignore
            except (AttributeError, TypeError):
                pass
        return cls.__instance__


singleton = metaclass(Singleton)


class Uninstantiable(type):
    """A metaclass that disallows instantiation."""

    def __call__(cls, *args, **kwargs):
        # type: (*Any, **Any) -> None
        """Do not allow the class to be instantiated."""
        raise TypeError('Type {} cannot be instantiated.'.format(cls.__name__))



# -*- coding: utf-8 -*-
"""A module containing base objects for creating typed classes.

Classes:
    BaseStrictObject: An object that asserts all annotated attributes are of
        the correct type.
    StrictObject: A derivative of BaseStrictObject that implements the default
        comparison operators and hash.
    BaseObject: An object that coerces all annotated attributes to the
        correct type.
    Object: A derivative of BaseObject that implements the default
        comparison operators and hash.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import inspect
import re
import tokenize
import types

import six
from typingplus import (  # noqa: F401 pylint: disable=unused-import
    cast,
    get_type_hints,
    is_instance,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
)

from .types import NoneType  # pylint: disable=redefined-builtin


__all__ = (
    'BaseStrictObject',
    'StrictObject',
    'BaseObject',
    'Object'
)


_T = TypeVar('_T')


def _get_type_name(type_):
    # type: (type) -> str
    """Return a displayable name for the type.

    Args:
        type_: A class object.

    Returns:
        A string value describing the class name that can be used in a natural
        language sentence.
    """
    name = repr(type_)
    if name.startswith('<'):
        name = getattr(type_, '__qualname__', getattr(type_, '__name__', ''))
    return name.rsplit('.', 1)[-1] or repr(type_)


def _get_class_frame_source(class_name):
    # type: (str) -> Optional[str]
    """Return the source code for a class by checking the frame stack.

    This is necessary because it is not possible to get the source of a class
    being created by a metaclass directly.

    Args:
        class_name: The class to look for on the stack.

    Returns:
        The source code for the requested class if the class was found and the
        source was accessible.
    """
    for frame_info in inspect.stack():
        try:
            with open(frame_info[1]) as fp:
                src = ''.join(
                    fp.readlines()[frame_info[2] - 1:])
        except IOError:
            continue
        if re.search(r'\bclass\b\s+\b{}\b'.format(class_name), src):
            reader = six.StringIO(src).readline
            tokens = tokenize.generate_tokens(reader)
            source_tokens = []
            indent_level = 0
            base_indent_level = 0
            has_base_level = False
            for token, value, _, _, _ in tokens:  # type: ignore
                source_tokens.append((token, value))
                if token == tokenize.INDENT:
                    indent_level += 1
                elif token == tokenize.DEDENT:
                    indent_level -= 1
                    if has_base_level and indent_level <= base_indent_level:
                        return (
                            tokenize.untokenize(source_tokens),
                            frame_info[0].f_globals,
                            frame_info[0].f_locals
                        )
                elif not has_base_level:
                    has_base_level = True
                    base_indent_level = indent_level
    raise TypeError(
        'Unable to retrieve source for class "{}"'.format(class_name))


def _is_propertyable(names,  # type: List[str]
                     attrs,  # type: Dict[str, Any]
                     annotations,  # type: Dict[str, type]
                     attr  # Dict[str, Any]
                     ):
    # type: (...) -> bool
    """Determine if an attribute can be replaced with a property.

    Args:
        names: The complete list of all attribute names for the class.
        attrs: The attribute dict returned by __prepare__.
        annotations: A mapping of all defined annotations for the class.
        attr: The attribute to test.

    Returns:
        True if the attribute can be replaced with a property; else False.
    """
    return (attr in annotations and
            not attr.startswith('_') and
            not attr.isupper() and
            '__{}'.format(attr) not in names and
            not isinstance(getattr(attrs, attr, None), types.MethodType))


def _create_typed_object_meta(get_fset):
    # type: (Callable[[str, str, Type[_T]], Callable[[_T], None]]) -> type
    """Create a metaclass for typed objects.

    Args:
        get_fset: A function that takes three parameters: the name of an
            attribute, the name of the private attribute that holds the
            property data, and a type. This function must an object method that
            accepts a value.

    Returns:
        A metaclass that reads annotations from a class definition and creates
        properties for annotated, public, non-constant, non-method attributes
        that will guarantee the type of the stored value matches the
        annotation.
    """
    def _get_fget(attr, private_attr, type_):
        # type: (str, str, Type[_T]) -> Callable[[], Any]
        """Create a property getter method for an attribute.

        Args:
            attr: The name of the attribute that will be retrieved.
            private_attr: The name of the attribute that will store any data
                related to the attribute.
            type_: The annotated type defining what values can be stored in the
                attribute.

        Returns:
            A function that takes self and retrieves the private attribute from
            self.
        """
        def _fget(self):
            # type: (...) -> Any
            """Get attribute from self without revealing the private name."""
            try:
                return vars(self)[private_attr]
            except AttributeError:
                raise AttributeError(
                    "'{}' object has no attribute '{}'".format(
                        _get_type_name(type_), attr))

        return _fget

    class _AnnotatedObjectMeta(type):
        """A metaclass that reads annotations from a class definition."""

        def __new__(mcs,  # type: Type[_AnnotatedObjectMeta]
                    name,  # type: str
                    bases,  # type: List[type]
                    attrs,  # type: Dict[str, Any]
                    **kwargs  # type: Dict[str, Any]
                    ):
            # type: (...) -> type
            """Create class objs that replaces annotated attrs with properties.

            Args:
                mcs: The class object being created.
                name: The name of the class to create.
                bases: The list of all base classes for the new class.
                attrs: The list of all attributes for the new class from the
                    definition.

            Returns:
                A new class instance with the expected base classes and
                attributes, but with annotated, public, non-constant,
                non-method attributes replaced by property objects that
                validate against the annotated type.
            """
            annotations = attrs.get('__annotations__', {})
            use_comment_type_hints = (
                not annotations and attrs.get('__module__') != __name__)
            if use_comment_type_hints:
                frame_source = _get_class_frame_source(name)
                annotations = get_type_hints(*frame_source)
            names = list(attrs) + list(annotations)
            typed_attrs = {}
            for attr in names:
                typed_attrs[attr] = attrs.get(attr)
                if _is_propertyable(names, attrs, annotations, attr):
                    private_attr = '__{}'.format(attr)
                    if attr in attrs:
                        typed_attrs[private_attr] = attrs[attr]
                    type_ = (
                        Optional[annotations[attr]] if
                        not use_comment_type_hints and
                        attr in attrs and
                        attrs[attr] is None
                        else annotations[attr]
                    )
                    typed_attrs[attr] = property(
                        _get_fget(attr, private_attr, type_),
                        get_fset(attr, private_attr, type_)
                    )
            properties = [attr for attr in annotations if _is_propertyable(
                          names, attrs, annotations, attr)]
            typed_attrs['_tp__typed_properties'] = properties
            typed_attrs['_tp__required_typed_properties'] = [
                attr for attr in properties if (
                    attr not in attrs or
                    attrs[attr] is None and use_comment_type_hints) and
                NoneType not in getattr(annotations[attr], '__args__', ())
            ]
            return super(_AnnotatedObjectMeta, mcs).__new__(  # type: ignore
                mcs, name, bases, typed_attrs, **kwargs)

    return _AnnotatedObjectMeta


def _strict_object_meta_fset(_, private_attr, type_):
    # type: (str, str, Type[_T]) -> Callable[[_T], None]
    """Create a property setter method for the attribute.

    Args:
        _: The name of the attribute to set. Unused.
        private_attr: The name of the attribute that will store any data
            related to the attribute.
        type_: The annotated type defining what values can be stored in the
            attribute.

    Returns:
        A method that takes self and a value and stores that value on self
        in the private attribute iff the value is an instance of type_.
    """
    def _fset(self,
              value  # type: Any
              ):
        # type: (...) -> None
        """Set the value on self iff the value is an instance of type_.

        Args:
            value: The value to set.

        Raises:
            TypeError: Raised when the value is not an instance of type_.
        """
        rtype = type_
        if isinstance(type_, TypeVar):
            type_map = dict(zip(
                self.__parameters__, self.__orig_class__.__args__))
            rtype = type_map[type_]
        if not is_instance(value, rtype):
            raise TypeError(
                'Cannot assign value of type {} to attribute of type {}.'
                .format(
                    _get_type_name(type(value)),
                    _get_type_name(rtype)))
        vars(self)[private_attr] = value

    return _fset


_StrictObjectMeta = _create_typed_object_meta(_strict_object_meta_fset)


def _object_meta_fset(_, private_attr, type_):
    # type: (str, str, Type[_T]) -> Callable[[_T], None]
    """Create a property setter method for the attribute.

    Args:
        _: The name of the attribute to set. Unused.
        private_attr: The name of the attribute that will store any data
            related to the attribute.
        type_: The annotated type defining what values can be stored in the
            attribute.

    Returns:
        A method that takes self and a value and stores that value on self
        in the private attribute if the value is not an instance of type_
        and cannot be cast into type_.
    """
    def _fset(self,
              value  # type: Any
              ):
        # type: (...) -> None
        """Set the value on self and coerce it to type_ if necessary.

        Args:
            value: The value to set.

        Raises:
            TypeError: Raised when the value is not an instance of type_
                and cannot be cast into a compatible object of type_.
        """
        rtype = type_
        if isinstance(type_, TypeVar):
            type_map = dict(zip(
                self.__parameters__, self.__orig_class__.__args__))
            rtype = type_map[type_]
        vars(self)[private_attr] = cast(rtype, value)

    return _fset


_ObjectMeta = _create_typed_object_meta(_object_meta_fset)


class _BaseAnnotatedObject(object):
    """A base class that looks for class attributes to create __init__."""

    def __init__(self, *args, **kwargs):
        """Set all attributes according to their annotation status."""
        super(_BaseAnnotatedObject, self).__init__()
        properties = self._tp__typed_properties
        required = self._tp__required_typed_properties
        positionals = zip(properties, args)
        for attr, value in positionals:
            if attr in kwargs:
                raise TypeError(
                    "__init__() got multiple values for argument '{}'"
                    .format(attr))
            kwargs[attr] = value
        missing = [attr for attr in required if attr not in kwargs]
        if missing:
            num_missing = len(missing)
            if num_missing > 1:
                args = ', '.join("'{}'".format(m) for m in missing[:-1])
                if num_missing > 2:
                    args += ','
                args += " and '{}'".format(missing[-1])
            else:
                args = "'{}'".format(missing[0])
            raise TypeError(
                '__init__() missing {} required argument{}: {}'.format(
                    num_missing, 's' if num_missing > 1 else '', args))
        for attr, value in six.iteritems(kwargs):
            if attr in properties:
                setattr(self, attr, value)

    def __repr__(self):
        # type: () -> str
        """Return a Python readable representation of the class."""
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(
                '{}={}'.format(attr_name, repr(getattr(self, attr_name)))
                for attr_name in
                self._tp__typed_properties))  # type: ignore


class _AnnotatedObjectComparisonMixin(object):
    """A mixin to add comparisons to classes made by _AnnotatedObjectMeta."""

    def _tp__get_typed_properties(self):
        """Return a tuple of typed attrs that can be used for comparisons.

        Raises:
            NotImplementedError: Raised if this class was mixed into a class
                that was not created by _AnnotatedObjectMeta.
        """
        try:
            return tuple(getattr(self, p) for p in
                         self._tp__typed_properties)
        except AttributeError:
            raise NotImplementedError

    def __eq__(self, other):
        """Test if two objects of the same base class are equal.

        If the objects are not of the same class, Python will default to
        comparison-by-ID.

        Args:
            other: The object to compare for equality.

        Returns:
            True if the objects are equal; else False.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return (self._tp__get_typed_properties() ==
                other._tp__get_typed_properties())

    def __ne__(self, other):
        """Test if two objects of the same class are not equal.

        If the objects are not of the same class, Python will default to
        comparison-by-ID.

        Args:
            other: The object to compare for non-equality.

        Returns:
            True if the objects are not equal; else False.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return not self == other

    def __lt__(self, other):
        """Test if self is less than an object of the same class.

        Args:
            other: The object to compare against.

        Returns:
            True if self is less than other; else False.

        Raises:
            TypeError: Raised if the objects are not of the same class.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return (self._tp__get_typed_properties() <
                other._tp__get_typed_properties())

    def __le__(self, other):
        """Test if self is less than or equal an object of the same class.

        Args:
            other: The object to compare against.

        Returns:
            True if self is less than or equal other; else False.

        Raises:
            TypeError: Raised if the objects are not of the same class.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return self == other or self < other

    def __gt__(self, other):
        """Test if self is greater than an object of the same class.

        Args:
            other: The object to compare against.

        Returns:
            True if self is greater than other; else False.

        Raises:
            TypeError: Raised if the objects are not of the same class.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return not self <= other

    def __ge__(self, other):
        """Test if self is greater than or equal an object of the same class.

        Args:
            other: The object to compare against.

        Returns:
            True if self is greater than or equal to other; else False.

        Raises:
            TypeError: Raised if the objects are not of the same class.
        """
        if other.__class__ is not self.__class__:
            return NotImplemented
        return not self < other

    def __hash__(self):
        """Generate a hash for the object based on the annotated attrs."""
        return hash(self._tp__get_typed_properties())


@six.add_metaclass(_StrictObjectMeta)  # type: ignore
class BaseStrictObject(_BaseAnnotatedObject):
    """A base class to create instance attrs for annotated class attrs.

    For every class attribute that is annotated, public, and not constant in
    the subclasses, this base class will generate property objects for the
    the instances that will enforce the type of the value set.

    If the subclass does not define __init__, a default implementation will be
    generated that takes all of the annotated, public, non-constant attributes
    as parameters. If an annotated attribute is not defined, it will be
    required in __init__.

    >>> from typet import BaseStrictObject
    >>> class Point(BaseStrictObject):
    ...     x: int
    ...     y: int
    ...
    ...
    >>> p = Point(0, 0)
    >>> p.x
    0
    >>> p.x = '0'
    Traceback (most recent call last):
        ...
    TypeError: Cannot assign value of type str to attribute of type int.
    """


class StrictObject(BaseStrictObject, _AnnotatedObjectComparisonMixin):
    """A base class to create instance attrs for annotated class attrs.

    For every class attribute that is annotated, public, and not constant in
    the subclasses, this base class will generate property objects for the
    the instances that will enforce the type of the value set.

    If the subclass does not define __init__, a default implementation will be
    generated that takes all of the annotated, public, non-constant attributes
    as parameters. If an annotated attribute is not defined, it will be
    required in __init__.

    >>> from typet import StrictObject
    >>> class Point(StrictObject):
    ...     x: int
    ...     y: int
    ...
    ...
    >>> p = Point(0, 0)
    >>> p.x
    0
    >>> p.x = '0'
    Traceback (most recent call last):
        ...
    TypeError: Cannot assign value of type str to attribute of type int.
    >>> p2 = Point(2, 2)
    >>> p < p2
    True
    >>> p > p2
    False
    """


@six.add_metaclass(_ObjectMeta)  # type: ignore
class BaseObject(_BaseAnnotatedObject):
    """A base class to create instance attrs for annotated class attrs.

    For every class attribute that is annotated, public, and not constant in
    the subclasses, this base class will generate property objects for the
    the instances that will enforce the type of the value set by attempting to
    cast the given value to the set type.

    If the subclass does not define __init__, a default implementation will be
    generated that takes all of the annotated, public, non-constant attributes
    as parameters. If an annotated attribute is not defined, it will be
    required in __init__.

    Additionally, this class implements basic comparison operators and the hash
    function.

    >>> from typet import BaseObject
    >>> class Point(BaseObject):
    ...     x: int
    ...     y: int
    ...
    ...
    >>> p = Point(0, 0)
    >>> p.x
    0
    >>> p.x = '5'
    >>> p.x
    5
    >>> p.x = 'five'
    Traceback (most recent call last):
        ...
    TypeError: Cannot convert 'five' to int.
    """


class Object(BaseObject, _AnnotatedObjectComparisonMixin):
    """A base class to create instance attrs for annotated class attrs.

    For every class attribute that is annotated, public, and not constant in
    the subclasses, this base class will generate property objects for the
    the instances that will enforce the type of the value set by attempting to
    cast the given value to the set type.

    If the subclass does not define __init__, a default implementation will be
    generated that takes all of the annotated, public, non-constant attributes
    as parameters. If an annotated attribute is not defined, it will be
    required in __init__.

    Additionally, this class implements basic comparison operators and the hash
    function.

    >>> from typet import Object
    >>> class Point(Object):
    ...     x: int
    ...     y: int
    ...
    ...
    >>> p = Point(0, 0)
    >>> p.x
    0
    >>> p.x = '5'
    >>> p.x
    5
    >>> p.x = 'five'
    Traceback (most recent call last):
        ...
    TypeError: Cannot convert 'five' to int.
    >>> p2 = Point(2, 2)
    >>> p < p2
    True
    >>> p > p2
    False
    """



# -*- coding: utf-8 -*-
"""Tests for py3.6+ annotation type hint casting."""

from __future__ import unicode_literals

import pytest

import typingplus as typing
import typet


def test_strict_object():
    """Simple test to verify basic StrictObject functionality."""
    class X(typet.StrictObject):
        x: str
    x = X('initial')
    x.x = 'hello'
    assert isinstance(x.x, str)
    assert x.x == 'hello'
    with pytest.raises(TypeError):
        x.x = 5


def test_object():
    """Simple test to verify basic Object functionality."""
    class X(typet.Object):
        x: str = None
    x = X()
    x.x = 5
    assert isinstance(x.x, str)
    assert x.x == '5'


def test_object_failure():
    """Simple test to verify basic Object failure functionality."""
    class X(typet.Object):
        x: int = None
    x = X()
    x.x = None
    with pytest.raises(TypeError):
        x.x = 'not an integer'


def test_optional_unassigned():
    """Verify that an unassigned Optional attribute is optional in __init__."""
    class X(typet.Object):
        x: typing.Optional[int]
    X()



# -*- coding: utf-8 -*-
"""Tests for annotation type hint casting."""

import os.path
import uuid

import pytest
import six

from typingplus import (  # noqa: F401 pylint: disable=unused-import
    is_instance,
    Any,
    Optional,
)
import typet


def test_bounded_type():
    """Test the bounded type object."""
    with pytest.raises(TypeError):
        BoundedInt = typet.Bounded[int]
    with pytest.raises(TypeError):
        BoundedInt = typet.Bounded[int, 10:20, lambda x: x, None]
    BoundedInt = typet.Bounded[int, 10:20]
    with pytest.raises(ValueError):
        BoundedInt(5)
    assert BoundedInt(10) == 10
    assert BoundedInt(15) == 15
    assert BoundedInt(20) == 20
    with pytest.raises(ValueError):
        BoundedInt(25)
    BoundedStr = typet.Bounded[str, 1:5, len]
    with pytest.raises(ValueError):
        BoundedStr('')
    assert BoundedStr('abc') == 'abc'
    with pytest.raises(ValueError):
        BoundedStr('abcdef')
    assert str(BoundedInt) == 'typet.validation.Bounded[int, 10:20]'
    assert typet.Bounded[Any, 10:20](15) == 15
    assert typet.Bounded['int', 20](15) == 15
    assert typet.Bounded['int', 10:](15) == 15


def test_length_type():
    """Test the bounded length type object."""
    with pytest.raises(TypeError):
        LengthBoundedStr = typet.Length[str]
    with pytest.raises(TypeError):
        LengthBoundedStr = typet.Length[str, 10:20, lambda x: x]
    LengthBoundedStr = typet.Length[str, 1:5]
    with pytest.raises(ValueError):
        LengthBoundedStr('')
    assert LengthBoundedStr('a') == 'a'
    assert LengthBoundedStr('abcde') == 'abcde'
    with pytest.raises(ValueError):
        LengthBoundedStr('abcdef')
    LengthBoundedList = typet.Length[list, 1:1]
    with pytest.raises(ValueError):
        LengthBoundedList([])
    assert LengthBoundedList([1]) == [1]
    with pytest.raises(ValueError):
        LengthBoundedList([1, 2])
    assert str(LengthBoundedStr) == 'typet.validation.Length[str, 1:5]'
    assert typet.Length[Any, 1:5]('abc') == 'abc'
    assert typet.Length['str', 20]('abc') == 'abc'


def test_string_type():
    """Test the bounded string type object."""
    with pytest.raises(TypeError):
        BoundedStr = typet.String[10:20, lambda x: x]
    BoundedStr = typet.String[1:5]
    with pytest.raises(ValueError):
        BoundedStr('')
    assert BoundedStr('a') == 'a'
    assert BoundedStr('abcde') == 'abcde'
    with pytest.raises(ValueError):
        BoundedStr('abcdef')
    assert str(BoundedStr) == 'typet.validation.String[1:5]'
    assert typet.String('hello') == 'hello'


def test_validation_type():
    """Test that the validation type validates content."""
    ValidFile = typet.Valid[os.path.isfile]
    assert ValidFile(__file__) == __file__
    with pytest.raises(TypeError):
        typet.Valid[int, int, int]


def test_path_types():
    """Test that the supplied path validation paths work."""
    assert str(typet.File(__file__))== __file__
    with pytest.raises(ValueError):
        typet.File(str(uuid.uuid4()))
    assert str(typet.Dir(os.path.dirname(__file__))) == os.path.dirname(
        __file__)
    with pytest.raises(ValueError):
        typet.Dir(str(uuid.uuid4()))
    assert str(typet.ExistingPath(__file__)) == __file__
    assert str(typet.ExistingPath(
        os.path.dirname(__file__))) == os.path.dirname(__file__)
    with pytest.raises(ValueError):
        typet.ExistingPath(str(uuid.uuid4()))


def test_none_type():
    """Verify that NoneType is type(None)."""
    assert typet.NoneType is type(None)


def test_singleton():
    """Test that a singleton only allows a single instance of a class."""
    @six.add_metaclass(typet.Singleton)
    class TestClass(object):
        pass

    assert TestClass() is TestClass()


def test_uninstantiable():
    """Test that an uninstantiable class cannot be instantiated."""
    @six.add_metaclass(typet.Uninstantiable)
    class TestClass(object):
        pass

    with pytest.raises(TypeError):
        TestClass()


def test_isinstance():
    """Test that instances of sliced type are instances of validation type."""
    Age = typet.Bounded[int, 0:150]
    assert isinstance(25, Age) is True
    assert isinstance(-5, Age) is False
    assert isinstance(200, Age) is False
    assert isinstance('not an int', Age) is False


def test_strict_object():
    """Simple test to verify basic StrictObject functionality."""
    class X(typet.StrictObject):
        x = None  # type: str
    x = X('initial')
    x.x = 'hello'
    assert is_instance(x.x, str)
    assert x.x == 'hello'
    with pytest.raises(TypeError):
        x.x = 5


def test_object():
    """Simple test to verify basic Object functionality."""
    class X(typet.Object):
        x = None  # type: Optional[str]
    x = X()
    x.x = 5
    assert is_instance(x.x, str)
    assert x.x == '5'


def test_object_comments():
    """Simple test to verify basic Object functionality with comment hints."""
    class X(typet.Object):
        x = None  # type: str
    with pytest.raises(TypeError):
        X()
    x = X(5)
    assert is_instance(x.x, str)
    assert x.x == '5'


def test_object_failure():
    """Simple test to verify basic Object failure functionality."""
    class X(typet.Object):
        x = None  # type: Optional[int]
    x = X()
    x.x = None
    with pytest.raises(TypeError):
        x.x = 'not an integer'



# -*- coding: utf-8 -*-
# pragma pylint: disable=bad-mcs-method-argument,bad-mcs-classmethod-argument
"""A module for handling with typing and type hints.

Classes:
    Bounded: A sliceable subclass of any class that raises a ValueError if the
        initialization value is out of bounds.
    Length: A sliceable subclass of any class that implements __len__ that
        raises a ValueError if the length of the initialization value is out of
        bounds.
    Valid: A sliceable type that validates any assigned value against a
        validation function.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import six
from typingplus import (  # noqa: F401 pylint: disable=unused-import
    eval_type,
    Any,
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union
)

from .meta import Uninstantiable


_T = TypeVar('_T')

if six.PY3:
    unicode = str  # pylint: disable=redefined-builtin
_STR_TYPE = unicode


class _ValidationMeta(type):
    """A metaclass that returns handles custom type checks."""

    __class_repr__ = None  # type: Optional[str]

    def __repr__(cls):
        # type: () -> str
        """Return a custom string for the type repr if defined."""
        if cls.__class_repr__:
            return cls.__class_repr__
        return super(_ValidationMeta, cls).__repr__()

    def __instancecheck__(cls, other):
        # type: (Any) -> bool
        """Determine if an instance is of the sliced type and within bounds.

        Args:
            other: The instance to test.

        Returns:
            True if the object is both of the same type as sliced by the
            created class as well as within the bounds defined by the class.
        """
        try:
            return bool(isinstance(other, cls.__type__) and  # type: ignore
                        cls(other))
        except ValueError:
            return False


class _BoundedMeta(Uninstantiable):
    """A metaclass that adds slicing to a class that creates new classes."""

    def __getitem__(cls, args):
        # type: (Union[Tuple[_T, Any], Tuple[_T, Any, Callable]]) -> type
        """Create a new subclass of a type bounded by the arguments.

        If a callable is passed as the third argument of the slice, it will be
        used as the comparison function for the boundaries.

        Args:
            args: A tuple with two or three parameters: a type, a slice
                representing the minimum and maximum lengths allowed for values
                of that type and, optionally, a function to use on values
                before comparing against the bounds.
        """
        type_, bound, keyfunc = cls._get_args(args)
        keyfunc_name = cls._get_fullname(keyfunc)
        identity = cls._identity
        BaseClass, MetaClass = cls._get_bases(type_)
        instantiate = cls._instantiate

        @six.add_metaclass(MetaClass)  # type: ignore
        class _BoundedSubclass(BaseClass):  # type: ignore
            """A subclass of type_ or object, bounded by a slice."""

            def __new__(cls, __value, *args, **kwargs):
                # type: (Type[_BoundedSubclass], Any, *Any, **Any) -> type
                """Return __value cast to _T.

                Any additional arguments are passed as-is to the constructor.

                Args:
                    __value: A value that can be converted to type _T.
                    args: Any additional positional arguments passed to the
                        constructor.
                    kwargs: Any additional keyword arguments passed to the
                        constructor.
                """
                instance = instantiate(
                    BaseClass, type_, __value, *args, **kwargs)
                cmp_val = keyfunc(instance)
                if bound.start is not None or bound.stop is not None:
                    if bound.start is not None and cmp_val < bound.start:
                        if keyfunc is not identity:
                            raise ValueError(
                                'The value of {}({}) [{}] is below the minimum'
                                ' allowed value of {}.'.format(
                                    keyfunc_name, repr(__value), repr(cmp_val),
                                    bound.start))
                        raise ValueError(
                            'The value {} is below the minimum allowed value '
                            'of {}.'.format(repr(__value), bound.start))
                    if bound.stop is not None and cmp_val > bound.stop:
                        if keyfunc is not identity:
                            raise ValueError(
                                'The value of {}({}) [{}] is above the maximum'
                                ' allowed value of {}.'.format(
                                    keyfunc_name, repr(__value), repr(cmp_val),
                                    bound.stop))
                        raise ValueError(
                            'The value {} is above the maximum allowed value '
                            'of {}.'.format(repr(__value), bound.stop))
                elif not cmp_val:
                    raise ValueError(
                        '{}({}) is False'.format(keyfunc_name, repr(instance)))
                return instance

        _BoundedSubclass.__type__ = type_
        _BoundedSubclass.__class_repr__ = cls._get_class_repr(
            type_, bound, keyfunc, keyfunc_name)
        return _BoundedSubclass

    @staticmethod
    def _get_bases(type_):
        # type: (type) -> Tuple[type, type]
        """Get the base and meta classes to use in creating a subclass.

        Args:
            type_: The type to subclass.

        Returns:
            A tuple containing two values: a base class, and a metaclass.
        """
        try:
            class _(type_):  # type: ignore
                """Check if type_ is subclassable."""
            BaseClass = type_
        except TypeError:
            BaseClass = object

        class MetaClass(_ValidationMeta, BaseClass.__class__):  # type: ignore
            """Use the type_ meta and include base validation functionality."""

        return BaseClass, MetaClass

    @staticmethod
    def _instantiate(class_, type_, __value, *args, **kwargs):
        """Instantiate the object if possible.

        Args:
            class_: The class to instantiate.
            type_: The the class is uninstantiable, attempt to cast to a base
                type.
            __value: The value to return if the class and type are
                uninstantiable.
            *args: The positional arguments to pass to the class.
            **kwargs: The keyword arguments to pass to the class.

        Returns:
            The class or base type instantiated using the arguments. If it is
            not possible to instantiate either, returns __value.
        """
        try:
            return class_(__value, *args, **kwargs)
        except TypeError:
            try:
                return type_(__value, *args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                return __value

    def _get_class_repr(cls, type_, bound, keyfunc, keyfunc_name):
        # type: (Any, slice, Callable, str) -> str
        """Return a class representation using the slice parameters.

        Args:
            type_: The type the class was sliced with.
            bound: The boundaries specified for the values of type_.
            keyfunc: The comparison function used to check the value
                boundaries.
            keyfunc_name: The name of keyfunc.

        Returns:
            A string representing the class.
        """
        if keyfunc is not cls._default:
            return '{}.{}[{}, {}, {}]'.format(
                cls.__module__, cls.__name__, cls._get_fullname(type_),
                cls._get_bound_repr(bound), keyfunc_name)
        return '{}.{}[{}, {}]'.format(
            cls.__module__, cls.__name__, cls._get_fullname(type_),
            cls._get_bound_repr(bound))

    def _get_args(cls, args):
        # type: (tuple) -> Tuple[Any, slice, Callable]
        """Return the parameters necessary to check type boundaries.

        Args:
            args: A tuple with two or three elements: a type, a slice
                representing the minimum and maximum lengths allowed for values
                of that type and, optionally, a function to use on values
                before comparing against the bounds.

        Returns:
            A tuple with three elements: a type, a slice, and a function to
            apply to objects of the given type. If no function was specified,
            it returns the identity function.
        """
        if not isinstance(args, tuple):
            raise TypeError(
                '{}[...] takes two or three arguments.'.format(cls.__name__))
        elif len(args) == 2:
            type_, bound = args
            keyfunc = cls._identity
        elif len(args) == 3:
            type_, bound, keyfunc = args
        else:
            raise TypeError(
                'Too many parameters given to {}[...]'.format(cls.__name__))
        if not isinstance(bound, slice):
            bound = slice(bound)
        return eval_type(type_), bound, keyfunc

    @staticmethod
    def _get_bound_repr(bound):
        # type: (slice) -> str
        """Return a string representation of a boundary slice.

        Args:
            bound: A slice object.

        Returns:
            A string representing the slice.
        """
        return '{}:{}'.format(bound.start or '', bound.stop or '')

    @staticmethod
    def _identity(obj):
        # type: (Any) -> Any
        """Return the given object.

        Args:
            obj: An object.

        Returns:
            The given object.
        """
        return obj

    _default = _identity  # type: Callable[[Any], Any]

    @staticmethod
    def _get_fullname(obj):
        # type: (Any) -> str
        """Get the full name of an object including the module.

        Args:
            obj: An object.

        Returns:
            The full class name of the object.
        """
        if not hasattr(obj, '__name__'):
            obj = obj.__class__
        if obj.__module__ in ('builtins', '__builtin__'):
            return obj.__name__
        return '{}.{}'.format(obj.__module__, obj.__name__)


@six.add_metaclass(_BoundedMeta)
class Bounded(object):
    """A type that creates a bounded version of a type when sliced.

    Bounded can be sliced with two or three elements: a type, a slice
    representing the minimum and maximum lengths allowed for values of that
    type and, optionally, a function to use on values before comparing against
    the bounds.

    >>> Bounded[int, 5:10](7)
    7
    >>> Bounded[int, 5:10](1)
    Traceback (most recent call last):
        ...
    ValueError: The value 1 is below the minimum allowed value of 5.
    >>> Bounded[int, 5:10](11)
    Traceback (most recent call last):
        ...
    ValueError: The value 11 is above the maximum allowed value of 10.
    >>> Bounded[str, 5:10, len]('abcde')
    'abcde'
    """


class _LengthBoundedMeta(_BoundedMeta):
    """A metaclass that bounds a type with the len function."""

    _default = len

    def _get_args(cls, args):
        # type: (tuple) -> Tuple[type, slice, Callable]
        """Return the parameters necessary to check type boundaries.

        Args:
            args: A tuple with two parameters: a type, and a slice representing
                the minimum and maximum lengths allowed for values of that
                type.

        Returns:
            A tuple with three parameters: a type, a slice, and the len
            function.
        """
        if not isinstance(args, tuple) or not len(args) == 2:
            raise TypeError(
                '{}[...] takes exactly two arguments.'.format(cls.__name__))
        return super(_LengthBoundedMeta, cls)._get_args(args + (len,))


@six.add_metaclass(_LengthBoundedMeta)
class Length(object):
    """A type that creates a length bounded version of a type when sliced.

    Length can be sliced with two parameters: a type, and a slice representing
    the minimum and maximum lengths allowed for values of that type.

    >>> Length[str, 5:10]('abcde')
    'abcde'
    >>> Length[str, 5:10]('abc')
    Traceback (most recent call last):
        ...
    ValueError: The value of len('abc') [3] is below the minimum ...
    >>> Length[str, 5:10]('abcdefghijk')
    Traceback (most recent call last):
        ...
    ValueError: The value of len('abcdefghijk') [11] is above the maximum ...
    """


class _ValidationBoundedMeta(_BoundedMeta):
    """A metaclass that binds a type to a validation method."""

    def _get_args(cls, args):
        # type: (tuple) -> Tuple[type, slice, Callable]
        """Return the parameters necessary to check type boundaries.

        Args:
            args: A tuple with one or two parameters: A type to cast the
                value passed, and a predicate function to use for bounds
                checking.

        Returns:
            A tuple with three parameters: a type, a slice, and the predicate
            function. If no type was passed in args, the type defaults to Any.
        """
        if isinstance(args, tuple):
            if not len(args) == 2:
                raise TypeError(
                    '{}[...] takes one or two argument.'.format(cls.__name__))
            return super(_ValidationBoundedMeta, cls)._get_args(
                (args[0], None, args[1]))
        return super(_ValidationBoundedMeta, cls)._get_args((Any, None, args))

    def _get_class_repr(cls, type_, bound, keyfunc, keyfunc_name):
        # type: (Any, slice, Callable, str) -> str
        """Return a class representation using the slice parameters.

        Args:
            type_: The type the class was sliced with.
            bound: The boundaries specified for the values of type_.
            keyfunc: The comparison function used to check the value
                boundaries.
            keyfunc_name: The name of keyfunc.

        Returns:
            A string representing the class.
        """
        if type_ is not Any:
            return '{}.{}[{}, {}]'.format(
                cls.__module__, cls.__name__, cls._get_fullname(type_),
                keyfunc_name)
        return '{}.{}[{}]'.format(
            cls.__module__, cls.__name__, keyfunc_name)


@six.add_metaclass(_ValidationBoundedMeta)
class Valid(object):
    """A type that creates a type that is validated against a function.

    Valid can be sliced with one or two parameters: an optional type to cast
    the given value to, and a validation method.
    """


class _StringMeta(_LengthBoundedMeta):
    """A metaclass that binds a string to a length bound."""

    def __call__(cls, *args, **kwargs):
        """Instantiate a string object."""
        return _STR_TYPE(*args, **kwargs)

    def __instancecheck__(self, other):
        # type: (Any) -> bool
        """Determine if an instance is of the string type.

        Args:
            other: The instance to test.

        Returns:
            True if the object is both of the same type as the String;
            otherwise, False,
        """
        return isinstance(other, _STR_TYPE)

    def _get_args(cls, args):
        # type: (tuple) -> Tuple[type, slice, Callable]
        """Return the parameters necessary to check type boundaries.

        Args:
            args: A slice representing the minimum and maximum lengths allowed
                for values of that string.

        Returns:
            A tuple with three parameters: a type, a slice, and the len
            function.
        """
        if isinstance(args, tuple):
            raise TypeError(
                '{}[...] takes exactly one argument.'.format(cls.__name__))
        return super(_StringMeta, cls)._get_args((_STR_TYPE, args))

    def _get_class_repr(cls, type_, bound, keyfunc, keyfunc_name):
        # type: (Any, slice, Callable, str) -> str
        """Return a class representation using the slice parameters.

        Args:
            type_: The type the class was sliced with. This will always be
                _STR_TYPE.
            bound: The boundaries specified for the values of type_.
            keyfunc: The comparison function used to check the value
                boundaries. This will always be builtins.len().
            keyfunc_name: The name of keyfunc. This will always be 'len'.

        Returns:
            A string representing the class.
        """
        return '{}.{}[{}]'.format(
            cls.__module__, cls.__name__, cls._get_bound_repr(bound))


@six.add_metaclass(_StringMeta)
class String(object):
    """A type that creates a length bounded version of a string when sliced.

    String represents a specific string type. On Python 2, this is 'unicode'
    and on Python 3 this is 'str'.

    String can be sliced with one parameter: a slice representing the minimum
    and maximum lengths allowed for values of the string.

    If String is not sliced, it will act as a constructor for the string type.

    >>> String[5:10]('abcde')
    'abcde'
    >>> String[5:10]('abc')
    Traceback (most recent call last):
        ...
    ValueError: The value of len('abc') [3] is below the minimum ...
    >>> String[5:10]('abcdefghijk')
    Traceback (most recent call last):
        ...
    ValueError: The value of len('abcdefghijk') [11] is above the maximum ...
    """


NonEmptyString = String[1:]



# -*- coding: utf-8 -*-
"""A module containing types representing file and directory states.

Classes:
    File: A type instance of Valid that validates that the value is a file.
    Dir: A type instance of Valid that validates that the value is a directory.
    Path: A type instance of Valid that expands a value to a path.
"""

from __future__ import unicode_literals

import os.path

from .validation import Valid

try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib  # type: ignore


def _valid(name, type_, predicate):
    new_type = Valid[type_, predicate]
    setattr(new_type, '__class_repr__', '{}.{}'.format(
        predicate.__module__, name))
    return new_type


def is_dir(path):
    """Determine if a Path or string is a directory on the file system."""
    try:
        return path.expanduser().absolute().is_dir()
    except AttributeError:
        return os.path.isdir(os.path.abspath(os.path.expanduser(str(path))))


def is_file(path):
    """Determine if a Path or string is a file on the file system."""
    try:
        return path.expanduser().absolute().is_file()
    except AttributeError:
        return os.path.isfile(os.path.abspath(os.path.expanduser(str(path))))


def exists(path):
    """Determine if a Path or string is an existing path on the file system."""
    try:
        return path.expanduser().absolute().exists()
    except AttributeError:
        return os.path.exists(os.path.abspath(os.path.expanduser(str(path))))


Dir = _valid('Dir', pathlib.Path, is_dir)
File = _valid('File', pathlib.Path, is_file)
ExistingPath = _valid('ExistingPath', pathlib.Path, exists)



# -*- coding: utf-8 -*-
"""A module containing basic types.

Classes:
    DefType: A union of function and method types.
    NoneType: An alias for type(None)
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import types

import typingplus as typing


__all__ = (
    'DefType',
    'NoneType',
)


DefType = typing.Union[types.FunctionType, types.MethodType]
NoneType = type(None)  # pylint: disable=redefined-builtin



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Install typet."""

from __future__ import unicode_literals

from setuptools import setup
from setuptools import find_packages


setup(
    name='typet',
    use_scm_version=True,
    description='A library of types that simplify working with typed Python.',
    long_description=open('README.rst').read(),
    author='Melissa Nuño',
    author_email='dangle@contains.io',
    url='https://github.com/contains-io/typet',
    keywords=['typing', 'schema' 'validation', 'types', 'annotation',
              'PEP 483', 'PEP 484', 'PEP 526'],
    license='MIT',
    packages=find_packages(exclude=['tests', 'docs']),
    install_requires=[
        'pathlib2',
        'typingplus >= 2.1.1, < 3'
    ],
    setup_requires=[
        'pytest-runner',
        'setuptools_scm',
    ],
    tests_require=['pytest >= 3.2'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
    ]
)
